#include <glib.h>
#include <stdlib.h>
#include <signal.h>

#include "tilda.h"
#include "tilda-window.h"
#include "tilda-terminal.h"
#include "tomboykeybinder.h"

DBusGConnection *dbus_connection;
GPtrArray *windows;

static void
tilda_initialize_dbus ()
{
	debug_enter  ();

	static const gchar service_name[] = "net.sourceforge.Tilda";
	GError *error = NULL;
	DBusGProxy *driver_proxy;
	guint request_ret;
	gboolean ret;

	// Initialize the DBus connection
	dbus_connection = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
	if (dbus_connection == NULL)
	{
		g_warning (_("Unable to connect to DBus: %s\n"), error->message);
		g_error_free (error);
		return;
	}

	// Register the service name
	driver_proxy = dbus_g_proxy_new_for_name (dbus_connection,
											  DBUS_SERVICE_DBUS,
											  DBUS_PATH_DBUS,
											  DBUS_INTERFACE_DBUS);

	ret = org_freedesktop_DBus_request_name (driver_proxy,
											 service_name,
											 DBUS_NAME_FLAG_DO_NOT_QUEUE,
											 &request_ret,
											 &error);

	if (!ret)
	{
		g_warning (_("Unable to communicate with DBus: %s\n"), error->message);
		g_error_free (error);
	}

	if (request_ret == DBUS_REQUEST_NAME_REPLY_EXISTS)
	{
		g_critical (_("There is already an instance of Tilda running\n"));
		exit (EXIT_FAILURE);
	}

	g_object_unref (driver_proxy);
}

static gint
tilda_find_next_free_window_number ()
{
	debug_enter  ();

	gint i, j;
	gboolean found;

	for (i=0; i<INT_MAX; ++i)
	{
		found = FALSE;

		for (j=0; j<windows->len; ++j)
		{
			TildaWindow *tw = g_ptr_array_index (windows, j);

			if (tw->number == i)
			{
				found = TRUE;
				break;
			}
		}

		if (!found)
			return i;
	}

	return 0;
}

static TildaWindow *
tilda_add_window ()
{
	debug_enter ();

	TildaWindow *ret;
	gint number;

	number = tilda_find_next_free_window_number ();
	ret = g_object_new (TILDA_TYPE_WINDOW, "number", number, NULL);

	g_ptr_array_add (windows, ret);

	debug_printf ("Adding window: 0x%x (number %d of %d)\n", ret, ret->number, windows->len-1);
	return ret;
}

void
tilda_del_window (gint number)
{
	debug_enter ();

	gint i;
	TildaWindow *win;

	for (i=0; i<windows->len; ++i)
	{
		win = g_ptr_array_index (windows, i);

		if (win->number == number)
		{
			debug_printf ("Deleting window 0x%x (number %d of %d)\n", win, win->number, windows->len-1);
			g_ptr_array_remove_index (windows, i);
			g_object_unref (G_OBJECT(win));

			if (windows->len == 0)
			{
				debug_printf ("No windows left, exiting...\n");
				gtk_main_quit ();
			}

			break;
		}
	}
}

static void
tilda_parse_command_line (gint argc, gchar *argv[])
{
	debug_enter ();

	gboolean version = FALSE;

	/* All of the various command-line options */
	const GOptionEntry cl_opts[] = {
		{ "version",			'V', 0, G_OPTION_ARG_NONE,		&version,			N_("Show version information"), NULL },
		{ NULL },
	};

	/* Set up the command-line parser */
	GError *error = NULL;
	GOptionContext *context = g_option_context_new (NULL);
	g_option_context_add_main_entries (context, cl_opts, NULL);
	g_option_context_add_group (context, gtk_get_option_group (TRUE));
	g_option_context_parse (context, &argc, &argv, &error);
	g_option_context_free (context);

	/* Check for unknown options, and give a nice message if there are some */
	if (error)
	{
		g_printerr (_("Error parsing command-line options: %s\n"), error->message);
		g_printerr (_("The command \"tilda --help\" will show all possible options\n"));
		g_error_free (error);
		exit (EXIT_FAILURE);
	}

	/* If we need to show the version, show it then exit normally */
	if (version)
	{
		g_print ("%s\n\n", TILDA_VERSION);

		g_print ("Copyright (c) 2005-2008 Tristan Sloughter (sloutri@iit.edu)\n");
		g_print ("Copyright (c) 2005-2008 Ira W. Snyder (tilda@irasnyder.com)\n\n");

		g_print ("This program comes with ABSOLUTELY NO WARRANTY.\n");
		g_print ("This is free software, and you are welcome to redistribute it\n");
		g_print ("under certain conditions. See the file COPYING for details.\n");

		exit (EXIT_SUCCESS);
	}
}

static void
tilda_termination_handler (gint signum)
{
	debug_enter  ();
	debug_printf ("signum: %d\n", signum);

	gtk_main_quit ();
}

/* Hook up all system signal handlers */
static void
tilda_signal_handlers_init ()
{
	struct sigaction sa;

	/* Hook up signal handlers */
	sa.sa_handler = tilda_termination_handler;
	sigemptyset (&sa.sa_mask);
	sa.sa_flags = 0;

	sigaction (SIGINT,  &sa, NULL);
	sigaction (SIGQUIT, &sa, NULL);
	sigaction (SIGABRT, &sa, NULL);
	sigaction (SIGTERM, &sa, NULL);

	/* SIGKILL cannot be caught according to sigaction(2) and signal(7) */
	/* sigaction (SIGKILL, &sa, NULL); */
}

int main (int argc, char *argv[])
{
	debug_enter ();

	TildaWindow *tw;

#if ENABLE_NLS
	/* Gettext Initialization */
	setlocale (LC_ALL, "");
	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);
#endif

	/* Parse the command-line options */
	tilda_parse_command_line (argc, argv);

	/* Initialize GTK+ (and the GObject system) */
	gtk_init (&argc, &argv);

	/* Hook up the signal handlers */
	tilda_signal_handlers_init ();

	/* Initialize the keybinder */
	tomboy_keybinder_init ();

	/* Start our connection to DBus */
	tilda_initialize_dbus ();

	/* Initialize the array of windows */
	windows = g_ptr_array_new ();

	/* Create a TildaWindow, run it, and exit when it does, basically.
	 *
	 * This is nothing like what the real main() will be, but it's
	 * a good start for testing and integration of more of TildaWindow
	 * and TildaTerminal. */
	tw = tilda_add_window ();

	debug_printf ("Starting gtk_main()!\n");
	gtk_main ();
	debug_printf ("Out of gtk_main(), going down\n");

	/* Free the pointer array we allocated earlier */
	g_ptr_array_free (windows, TRUE);

	return 0;
}

/* vim: set ts=4 sts=4 sw=4 noet tw=112: */
